import { Message, MessageContent, MessageFile, TextableChannel, TextChannel } from "eris";
import { GuildCases, ICaseDetails } from "../data/GuildCases";
import { CaseTypes } from "../data/CaseTypes";
import { Case } from "../data/entities/Case";
import moment from "moment-timezone";
import { CaseTypeColors } from "../data/CaseTypeColors";
import { ZeppelinPlugin } from "./ZeppelinPlugin";
import { GuildActions } from "../data/GuildActions";
import { GuildArchives } from "../data/GuildArchives";
import { IPluginOptions } from "knub";

interface ICasesPluginConfig {
  log_automatic_actions: boolean;
  case_log_channel: string;
}

export class CasesPlugin extends ZeppelinPlugin<ICasesPluginConfig> {
  public static pluginName = "cases";

  protected actions: GuildActions;
  protected cases: GuildCases;
  protected archives: GuildArchives;

  getDefaultOptions(): IPluginOptions<ICasesPluginConfig> {
    return {
      config: {
        log_automatic_actions: true,
        case_log_channel: null,
      },
    };
  }

  onLoad() {
    this.actions = GuildActions.getInstance(this.guildId);
    this.cases = GuildCases.getInstance(this.guildId);
    this.archives = GuildArchives.getInstance(this.guildId);

    this.actions.register("createCase", args => {
      return this.createCase(args);
    });

    this.actions.register("createCaseNote", args => {
      return this.createCaseNote(
        args.caseId,
        args.modId,
        args.note,
        args.automatic,
        args.postInCaseLog,
        args.noteDetails,
      );
    });

    this.actions.register("postCase", async args => {
      const embed = await this.getCaseEmbed(args.caseId);
      return (args.channel as TextableChannel).createMessage(embed);
    });
  }

  onUnload() {
    this.actions.unregister("createCase");
    this.actions.unregister("createCaseNote");
    this.actions.unregister("postCase");
  }

  protected resolveCaseId(caseOrCaseId: Case | number): number {
    return caseOrCaseId instanceof Case ? caseOrCaseId.id : caseOrCaseId;
  }

  /**
   * Creates a new case and, depending on config, posts it in the case log channel
   * @return {Number} The ID of the created case
   */
  public async createCase(opts: ICaseDetails): Promise<Case> {
    const user = this.bot.users.get(opts.userId);
    const userName = user ? `${user.username}#${user.discriminator}` : "Unknown#0000";

    const mod = this.bot.users.get(opts.modId);
    const modName = mod ? `${mod.username}#${mod.discriminator}` : "Unknown#0000";

    let ppName = null;
    if (opts.ppId) {
      const pp = this.bot.users.get(opts.ppId);
      ppName = pp ? `${pp.username}#${pp.discriminator}` : "Unknown#0000";
    }

    const createdCase = await this.cases.create({
      type: opts.type,
      user_id: opts.userId,
      user_name: userName,
      mod_id: opts.modId,
      mod_name: modName,
      audit_log_id: opts.auditLogId,
      pp_id: opts.ppId,
      pp_name: ppName,
    });

    if (opts.reason || (opts.noteDetails && opts.noteDetails.length)) {
      await this.createCaseNote(createdCase, opts.modId, opts.reason || "", opts.automatic, false, opts.noteDetails);
    }

    if (opts.extraNotes) {
      for (const extraNote of opts.extraNotes) {
        await this.createCaseNote(createdCase, opts.modId, extraNote, opts.automatic, false);
      }
    }

    const config = this.getConfig();

    if (
      config.case_log_channel &&
      (!opts.automatic || config.log_automatic_actions) &&
      opts.postInCaseLogOverride !== false
    ) {
      try {
        await this.postCaseToCaseLogChannel(createdCase);
      } catch (e) {} // tslint:disable-line
    }

    return createdCase;
  }

  /**
   * Adds a case note to an existing case and, depending on config, posts the updated case in the case log channel
   */
  public async createCaseNote(
    caseOrCaseId: Case | number,
    modId: string,
    body: string,
    automatic = false,
    postInCaseLogOverride = null,
    noteDetails: string[] = null,
  ): Promise<void> {
    const mod = this.bot.users.get(modId);
    const modName = mod ? `${mod.username}#${mod.discriminator}` : "Unknown#0000";

    const theCase = await this.cases.find(this.resolveCaseId(caseOrCaseId));
    if (!theCase) {
      this.throwPluginRuntimeError(`Unknown case ID: ${caseOrCaseId}`);
    }

    // Add note details to the beginning of the note
    if (noteDetails && noteDetails.length) {
      body = noteDetails.map(d => `__[${d}]__`).join(" ") + " " + body;
    }

    await this.cases.createNote(theCase.id, {
      mod_id: modId,
      mod_name: modName,
      body: body || "",
    });

    if (theCase.mod_id == null) {
      // If the case has no moderator information, assume the first one to add a note to it did the action
      await this.cases.update(theCase.id, {
        mod_id: modId,
        mod_name: modName,
      });
    }

    const archiveLinkMatch = body && body.match(/\/archives\/([a-zA-Z0-9\-]+)/);
    if (archiveLinkMatch) {
      const archiveId = archiveLinkMatch[1];
      this.archives.makePermanent(archiveId);
    }

    if ((!automatic || this.getConfig().log_automatic_actions) && postInCaseLogOverride !== false) {
      try {
        await this.postCaseToCaseLogChannel(theCase.id);
      } catch (e) {} // tslint:disable-line
    }
  }

  /**
   * Returns a Discord embed for the specified case
   */
  public async getCaseEmbed(caseOrCaseId: Case | number): Promise<MessageContent> {
    const theCase = await this.cases.with("notes").find(this.resolveCaseId(caseOrCaseId));
    if (!theCase) return null;

    const createdAt = moment(theCase.created_at);
    const actionTypeStr = CaseTypes[theCase.type].toUpperCase();

    const embed: any = {
      title: `${actionTypeStr} - Case #${theCase.case_number}`,
      footer: {
        text: `Case created at ${createdAt.format("YYYY-MM-DD [at] HH:mm")}`,
      },
      fields: [
        {
          name: "User",
          value: `${theCase.user_name}\n<@!${theCase.user_id}>`,
          inline: true,
        },
        {
          name: "Moderator",
          value: `${theCase.mod_name}\n<@!${theCase.mod_id}>`,
          inline: true,
        },
      ],
    };

    if (theCase.pp_id) {
      embed.fields[1].value += `\np.p. ${theCase.pp_name}\n<@!${theCase.pp_id}>`;
    }

    if (theCase.is_hidden) {
      embed.title += " (hidden)";
    }

    if (CaseTypeColors[theCase.type]) {
      embed.color = CaseTypeColors[theCase.type];
    }

    if (theCase.notes.length) {
      theCase.notes.forEach((note: any) => {
        const noteDate = moment(note.created_at);
        embed.fields.push({
          name: `${note.mod_name} at ${noteDate.format("YYYY-MM-DD [at] HH:mm")}:`,
          value: note.body,
        });
      });
    } else {
      embed.fields.push({
        name: "!!! THIS CASE HAS NO NOTES !!!",
        value: "\u200B",
      });
    }

    return { embed };
  }

  /**
   * A helper for posting to the case log channel.
   * Returns silently if the case log channel isn't specified or is invalid.
   */
  public postToCaseLogChannel(content: MessageContent, file: MessageFile = null): Promise<Message> {
    const caseLogChannelId = this.getConfig().case_log_channel;
    if (!caseLogChannelId) return;

    const caseLogChannel = this.guild.channels.get(caseLogChannelId);
    if (!caseLogChannel || !(caseLogChannel instanceof TextChannel)) return;

    return caseLogChannel.createMessage(content, file);
  }

  /**
   * A helper to post a case embed to the case log channel
   */
  public async postCaseToCaseLogChannel(caseOrCaseId: Case | number): Promise<Message> {
    const caseEmbed = await this.getCaseEmbed(caseOrCaseId);
    if (!caseEmbed) return;

    return this.postToCaseLogChannel(caseEmbed);
  }
}
